from save_to_db.exceptions import BulkItemOneToXDefaultError, WrongAlias
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestBasicUse(TestBase):
    
    ModelFieldTypes = None
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    @classmethod
    def setup_models(cls, aliased=False):
        cls.item_cls_manager.clear()
        
        dct_1 = {'model_cls': cls.ModelGeneralOne}
        dct_2 = {'model_cls': cls.ModelGeneralTwo}
        
        if aliased:
            dct_1.update({
                'aliases': {
                    'alias_1': 'two_1_1__f_integer',
                    'alias_2': 'two_x_x__one_1_1__f_integer',
                }
            })
            dct_2.update({
                'aliases': {
                    'alias_1__post': 'f_string',
                    'alias_2__post': 'one_x_1__alias_1',
                    'alias_3__post': 'one_x_x__two_1_x__alias_1__post',
                }
            })
        
        cls.ItemGeneralOne = type('ItemGeneralOne', (Item,), dct_1)
        cls.ItemGeneralTwo = type('ItemGeneralTwo', (Item,), dct_2)
        
        cls.ItemGeneralOne.complete_setup()
        cls.ItemGeneralTwo.complete_setup()
    
    
    def test_set_field(self):
        self.setup_models()

        #--- setting with correct keys ---
        item = self.ItemGeneralOne()
        item['f_integer'] = '10'
        item['two_1_1__f_string'] = 'one'
        item['two_1_1__one_1_x__f_string'] = 'two'
        item['two_1_1__one_x_x__two_x_1__f_string'] = 'three'
        bulk = item['two_1_1__one_x_x']
        bulk['f_integer'] = '20'
        bulk.gen(f_text='four')
        in_bulk_item = bulk.gen(f_text='five')
        # setting field by calling
        in_bulk_item(f_string='ITEM_CALL') 
        bulk(two_x_1__f_text='BULK_CALL')
        
        self.assertEqual(item['f_integer'], '10')
        self.assertEqual(item['two_1_1']['f_string'], 'one')
        self.assertEqual(item['two_1_1']['one_1_x']['f_string'], 'two')
        self.assertNotIn('two_x_1', item['two_1_1']['one_x_x'])
        self.assertEqual(item['two_1_1']['one_x_x']['two_x_1__f_string'],
                         'three')
        
        bulk = item['two_1_1__one_x_x']
        self.assertEqual(bulk['f_integer'], '20')
        self.assertEqual(bulk[0]['f_text'], 'four')
        
        self.assertEqual(bulk[1]['f_text'], 'five')
        
        self.assertEqual(bulk[1]['f_string'], 'ITEM_CALL')
        self.assertEqual(bulk['two_x_1__f_text'], 'BULK_CALL')
        
        #--- setting with incorrect keys ---
        item = self.ItemGeneralOne()
        with self.assertRaises(WrongAlias):
            item['wrong_key'] = 'value'
        
        bulk = self.ItemGeneralOne.Bulk()
        with self.assertRaises(WrongAlias):
            bulk['wrong_key'] = 'value'
            
        #--- setting one-to-x default value in bulk
        
        # not items temselves, can belong to different items
        bulk['two_1_1__f_integer'] = '10'
        bulk['two_1_x__f_integer'] = '10'
        bulk['two_x_x__one_1_x__f_integer'] = '10'
         
        # direct
        with self.assertRaises(BulkItemOneToXDefaultError):
            bulk['two_1_1'] = self.ItemGeneralTwo()
         
        with self.assertRaises(BulkItemOneToXDefaultError):
            bulk['two_1_x'] = self.ItemGeneralTwo()
         
        with self.assertRaises(BulkItemOneToXDefaultError):
            bulk['two_1_x'].gen()
         
        with self.assertRaises(BulkItemOneToXDefaultError):
            bulk['two_x_x__one_1_x'].gen()
        
        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        # aliased
        self.assertEqual(item['alias_1'], '1')
        self.assertEqual(item['alias_2'], '2')
        self.assertEqual(item['two_1_1__alias_1__post'], 'str-1')
        self.assertEqual(item['two_1_1__alias_2__post'], '3')
        self.assertEqual(item['two_1_1__alias_3__post'], '4')
        
        bulk = item['two_1_x']
        self.assertEqual(bulk['alias_1__post'], 'str-2')
        
        # direct
        self.assertEqual(item['two_1_1__f_integer'], '1')
        self.assertEqual(item['two_x_x__one_1_1__f_integer'], '2')
        self.assertEqual(item['two_1_1__f_string'], 'str-1')
        self.assertEqual(item['two_1_1__one_x_1__two_1_1__f_integer'], '3')
        self.assertEqual(item['two_1_1__one_x_x__two_1_x__f_string'], '4')
        
        bulk = item['two_1_x']
        self.assertEqual(bulk['f_string'], 'str-2')
    
    
    def test_aliases(self):
        self.setup_models(aliased=True)
        
        # set
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        # get
        self.assertEqual(item['alias_1'], '1')
        self.assertEqual(item['alias_2'], '2')
        self.assertEqual(item['two_1_1__alias_1__post'], 'str-1')
        self.assertEqual(item['two_1_1__alias_2__post'], '3')
        self.assertEqual(item['two_1_1__alias_3__post'], '4')
        self.assertEqual(bulk['alias_1__post'], 'str-2')
        
        # delete
        self.assertIn('alias_1', item)
        del item['alias_1']
        self.assertNotIn('alias_1', item)
        self.assertIn('alias_2', item)
        
        self.assertIn('two_1_1__alias_1__post', item)
        del item['two_1_1__alias_1__post']
        self.assertNotIn('two_1_1__alias_1__post', item)
        
    
    def test_del_field(self):
        self.setup_models()
        
        item = self.ItemGeneralOne()
        item['f_string'] = 'str-10'
        item['two_1_1__f_string'] = 'str-20'
        item['two_x_x__one_1_1__f_string'] = 'str-30'
        bulk = item['two_1_x']
        bulk['f_string'] = 'str-2'
        
        self.assertEqual(item['f_string'], 'str-10')
        del item['f_string']
        self.assertNotIn('f_string', item)
        
        self.assertEqual(item['two_1_1']['f_string'], 'str-20')
        del item['two_1_1__f_string']
        self.assertNotIn('f_string', item['two_1_1'])
        
        
        self.assertEqual(item['two_x_x']['one_1_1__f_string'], 'str-30')
        self.assertNotIn('one_1_1', item['two_x_x'])

        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        self.assertEqual(item['alias_1'], '1')
        del item['alias_1']
        self.assertNotIn('alias_1', item)
        
        self.assertEqual(item['alias_2'], '2')
        del item['alias_2']
        self.assertNotIn('alias_2', item)

        self.assertEqual(item['two_1_1__alias_1__post'], 'str-1')
        del item['two_1_1__alias_1__post']
        self.assertNotIn('two_1_1__alias_1__post', item)
        
        self.assertEqual(item['two_1_1__alias_2__post'], '3')
        del item['two_1_1__alias_2__post']
        self.assertNotIn('two_1_1__alias_2__post', item)
        
        self.assertEqual(item['two_1_1__alias_3__post'], '4')
        del item['two_1_1__alias_3__post']
        self.assertNotIn('two_1_1__alias_3__post', item)
        
        bulk = item['two_1_x']
        
        self.assertEqual(bulk['alias_1__post'], 'str-2')
        del bulk['alias_1__post']
        self.assertNotIn('alias_1__post', bulk)
    
    
    def test_contains(self):
        self.setup_models()
        item = self.ItemGeneralOne()
        item['f_integer'] = '1'
        item['two_1_x__f_integer'] = '2'
        bulk = item['two_1_x']
        bulk['one_x_x__f_integer'] = '3'
        
        self.assertIn('f_integer', item)
        self.assertIn('two_1_x__f_integer', item)
        self.assertIn('two_1_x__one_x_x__f_integer', item)  # from bulk
        
        self.assertNotIn('wrong_key', item)
        self.assertNotIn('wrong_key__f_integer', item)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', item)
        
        self.assertIn('f_integer', bulk)  # item
        self.assertIn('one_x_x__f_integer', bulk)
        
        self.assertNotIn('wrong_key', bulk)
        self.assertNotIn('wrong_key__f_integer', bulk)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', bulk)
        
        #--- aliased -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        self.assertIn('alias_1', item)
        self.assertIn('alias_2', item)
        self.assertIn('two_1_1__alias_1__post', item)
        self.assertIn('two_1_1__alias_2__post', item)
        self.assertIn('two_1_1__alias_3__post', item)
        
        self.assertNotIn('wrong_key', item)
        self.assertNotIn('wrong_key__f_integer', item)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', item)
        
        self.assertIn('alias_1__post', bulk)
        
        self.assertNotIn('wrong_key', bulk)
        self.assertNotIn('wrong_key__f_integer', bulk)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', bulk)
    
    
    def test_bulk_iter(self):
        self.setup_models()
        bulk = self.ItemGeneralOne.Bulk()
        items = [bulk.gen(f_integer=1),
                 bulk.gen(f_integer=2),
                 bulk.gen(f_integer=3),]
        items_in_bulk = list(bulk)
        self.assertEqual(items_in_bulk, items)
        
        for item_no, item_in_bulk in enumerate(bulk):
            self.assertIs(items[item_no], item_in_bulk)
            self.assertIs(items[item_no], bulk[item_no])
        
        self.assertEqual(item_no, len(items)-1)
        
        
    def test_bulk_item_slice(self):
        class ItemsGeneralOne(Item):
            model_cls = self.ModelGeneralOne
                
        class ItemsGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
                
        
        bulk = ItemsGeneralOne.Bulk()
        
        items = []
        for i in range(10):
            item = ItemsGeneralOne(f_integer=i)
            items.append(item)
            bulk.add(item)
        
        self.assertEqual(items[:5], bulk[:5])
        self.assertEqual(items[0::2], bulk[0::2])




from .utils.item_collection import ItemCollection


class ItemClsManager(ItemCollection):
    """ This class manages all known item classes
    (see :py:class:`~.item.Item` base class). Normally only one instance of
    this class is used.
    
    :param autogenerate: If `True` then getting a missing item by model class
        will generate a new item class with all parameters auto configured.
    """
    
    def __init__(self, autogenerate=False):
        super().__init__(collection_id=None)
        self.autogenerate = autogenerate

    
    def get_by_path(self, path, relative_to=None):
        """ Gets item classes using `path` (module path and class name separated
        by dot).
        
        :param path: Item path. Multiple initial dots allowed, e.g.
            `'...some_module.SomeClass'`.
        :param relative_to: In case `path` argument is relative, this arguments
            contains a path it is relative to.
        :return: List of item classes.
        """
        
        result = []
        # if path is relative or just a class name
        if path.startswith('.') or '.' not in path:
            
            if path.startswith('.'):
                if relative_to is None:
                    raise Exception('Relative path not allowed without setting '
                                    '`relative_to` argument.')
                stripped_path = path.lstrip('.')
                dot_count = len(path) - len(stripped_path)
                if dot_count > 1:
                    path_parts = relative_to.split('.')[:-(dot_count-1)]
                    relative_to = '.'.join(path_parts)
            else:
                stripped_path = path
            
            if relative_to:
                path = '{}.{}'.format(relative_to, stripped_path)      
        
        search_by_name = '.' not in path
        for item_cls in self.collection:
            # by name
            if search_by_name:
                if item_cls.__name__ == path:
                    result.append(item_cls)
                    continue
            
            # by path
            item_path = '{}.{}'.format(item_cls.__module__,
                                       item_cls.__name__)
            if item_path == path:
                result.append(item_cls)
                
        return result
    
    
    def __autogenerate_item_cls(self, model_cls):
        from .item import Item  # circular reference
        
        return Item._Auto(model_cls, metadata={
                    'autogenerated_item_cls': True,
                })
    
    
    def get_by_model_cls(self, model_cls, collection_id=None):
        """ Gets item classes using ORM model class.
        
        :param model_cls: An ORM model class that an item is using to persist
            data.
        :param collection_id: Scope ID of the item class.
        :return: List of item classes.
        """
        from .scope import Scope
        
        result = []
        if collection_id:
            scope = Scope.get_collection_by_id(collection_id)
        else:
            scope = None
        
        for item_cls in self.collection:
            if item_cls.model_cls is model_cls:
                if scope:
                    item_cls = scope[item_cls]
                result.append(item_cls)
        
        if not result and self.autogenerate:
            item_cls = self.__autogenerate_item_cls(model_cls)
            if scope:
                item_cls = scope[item_cls]
            result.append(item_cls)
            
        return result
    
    
    def autocomplete_item_classes(self):
        """ If `self.autogenerate` is `True` then auto generates missing item
        classes for models classes that are needed for already registered item
        classes via relations.
        """
        if not self.autogenerate:
            return
        
        # generating all missing items
        while True:
            regestry_copy = self.collection.copy()
            for item_cls in regestry_copy:
                item_cls.complete_setup()
                for relation in item_cls.relations.values():
                    relation['item_cls'].complete_setup()
            if regestry_copy == self.collection:
                break
        
    def __repr__(self):
        return '<{}()>'.format(self.__class__.__name__)


#: Normally only this instance of :py:class:`~ItemClsManager` is used.
item_cls_manager = ItemClsManager()



import inspect
from save_to_db.adapters import ADAPTERS


name = 'Save-to-DB'
short_description = ('Make it easy to store data from any source into a '
                     'database.')
release = '0.7.2a'


def print_info():
    to_print = ['='*80, '{}, release: {}'.format(name, release)]
    to_print.append(short_description)
    to_print.append('='*80)
    to_print.append('Number of ORM adapters: {}'.format(len(ADAPTERS)))
    for i, adapter_cls in enumerate(ADAPTERS):
        to_print.append('-'*80)
        to_print.append('{}. {}:'.format(i+1, adapter_cls.__name__))
        to_print.append(inspect.getdoc(adapter_cls))
    to_print.append('='*80)
    
    print('\n'.join(to_print))





from collections import defaultdict
from .merge_items import merge_items
from ..item_base import ItemBase
from save_to_db.exceptions import MultipleItemsMatch


def complete_item_structure(item):
    """ Takes an item, merges items that must point to same model (if allowed),
    checkes collisions and returns a structure like this:
    
        .. code-block:: Python
            
            {
                item_cls: [item_instance, ...],
                ...
            }
            
    Where keys are a subclasses of :py:class:`~.item.Item` and values are
    instances of that class.
    """
    top_item = item
    
    # item class to a list of instances
    structure = defaultdict(list)
    default_structure = defaultdict(list)
    
    __add_flat_item_to_structure(top_item, structure, default_structure)
    
    # merging items and checking collisions
    all_items = []
    for items in structure.values():
        all_items.extend(items)

    for item_cls in structure.keys():
        if item_cls.allow_merge_items:
            merge_items(top_item,
                        structure[item_cls], default_structure[item_cls],
                        all_items)
        else:
            __check_item_collisions(structure[item_cls])
     
    return structure


def __add_flat_item_to_structure(item, structure, default_structure,
                                 is_default_item=False):
    
    if is_default_item:
        store_to_structure = default_structure
    else:
        store_to_structure = structure
    
    if item.is_single_item():
        if item in store_to_structure[type(item)]:
            return
        store_to_structure[type(item)].append(item)
        
        for key in item.relations:
            if key not in item or item[key] is None:
                continue
            
            __add_flat_item_to_structure(
                item[key], store_to_structure, default_structure,
                is_default_item=is_default_item
            )
                        
    else:
        # defaults in bulk
        for key, default_item in item.data.items():
            if not isinstance(default_item, ItemBase):
                continue
            
            __add_flat_item_to_structure(
                default_item, store_to_structure, default_structure,
                is_default_item=True
            )
        
        # normal values in bulk
        for inner_item in item.bulk:
            __add_flat_item_to_structure(
                inner_item, store_to_structure, default_structure,
                is_default_item=is_default_item
            )


def __check_item_collisions(item_list):
    if not item_list:
        return
    
    # `item_list` must contain items of the same class
    getter_groups = item_list[0].getters
    
    if not getter_groups:
        return
    
    for left_item in item_list:
        for right_item in item_list:
            if left_item is right_item:
                continue
             
            for getters in getter_groups:
                same_by_group = True
                for key in getters:
                    if key not in left_item.data or key not in right_item.data \
                            or left_item[key] is None or \
                            left_item[key] != right_item[key]:
                        same_by_group = False
                        break
      
                     
                if same_by_group:
                    raise MultipleItemsMatch(None, left_item, right_item,
                                             getters)


from collections import defaultdict
from weakref import WeakSet


class Mapper(object):
    """ This class automatically sets up reverse relations for all items. """
    
    def __init__(self):
        # [collection_id][item_cls] = list_of_items
        self._map = defaultdict(lambda: defaultdict(WeakSet))
    
    
    def __contains__(self, item):
        return item in self.__get_item_set(item)
    
    
    def __get_item_set(self, item):
        collection_id = item.get_collection_id()
        item_cls = item.get_item_cls()
        return self._map[collection_id][item_cls]
    
    
    def __get_item_collection_set(self, item):
        collection_id = item.get_collection_id()
        return self._map[collection_id]
    
    
    def item_created(self, item):
        """ Called when an instance of
        :py:class:`~save_to_db.core.item_base.ItemBase` is created.
        
        :param item: An instance of created item.
        """
        self.__get_item_set(item).add(item)
    
    
    def item_value_set(self, item, key, old_value, new_value):
        """ Called when an item attribute is set, for example:
        
            .. code-block:: Python
                
                item[key] = new_value
        
        :param item: An item whose attribute is set.
        :param key: Attribute name of the `item`.
        :param old_value: Old attribute value.
        :param new_value: New attribute value.
        """
        if item.is_bulk_item():
            return  # default value was set
        
        if key not in item.relations:
            return
        
        relation = item.relations[key]
        reverse_key = relation['reverse_key']
        
        this_item, that_item_new = item, new_value
        that_item_old = old_value
        
        if not reverse_key:
                return
        
        if relation['relation_type'].is_one_to_one():
        
            # reverse for old
            if that_item_old and reverse_key in that_item_old.data:
                del that_item_old.data[reverse_key]
            
            # reverse for new
            if that_item_new is None:
                return
        
            that_item_data = that_item_new.data
            if reverse_key in that_item_data and \
                    that_item_data is this_item.data:
                return
            that_item_new.data[reverse_key] = this_item
        
        elif relation['relation_type'].is_one_to_many():
            
            # reverse for old
            if that_item_old:
                for item_in_bulk in that_item_old.bulk:
                    del item_in_bulk.data[reverse_key]
            
            # reverse for new
            if that_item_new:
                for item_in_bulk in that_item_new.bulk:
                    item_in_bulk.data[reverse_key] = this_item
        
        elif relation['relation_type'].is_many_to_one():
            
            # reverse for old item
            if that_item_old:
                if item in that_item_old[reverse_key].bulk:
                    that_item_old[reverse_key].bulk.remove(item)
            
            # reverse for new item
            if that_item_new:
                if item not in that_item_new[reverse_key].bulk:
                    that_item_new[reverse_key].bulk.append(item)
        
        elif relation['relation_type'].is_many_to_many():
            
            # reverse for old item
            if that_item_old:
                for item_in_bulk in that_item_old.bulk:
                    if item in item_in_bulk[reverse_key].bulk:
                        item_in_bulk[reverse_key].bulk.remove(item)
            
            # reverse for new item
            if that_item_new:
                for item_in_bulk in that_item_new.bulk:
                    if item not in item_in_bulk[reverse_key].bulk:
                        item_in_bulk[reverse_key].bulk.append(item)
        
    
    def item_value_deleted(self, item, key, old_value):
        """ Called when an item attribute is deleted, for example:
        
            .. code-block:: Python
                
                del item[key]
        
        :param item: An item whose attribute is deleted.
        :param key: Deleted attribute name.
        :param old_value: Deleted attribute value.
        """
        if item.is_bulk_item():
            return  # default value deleted
        
        if key not in item.relations or old_value is None:
            return
        
        relation = item.relations[key]
        if not relation['reverse_key']:
            return
        
        if relation['relation_type'].is_one_to_one():
            if relation['reverse_key'] in old_value:
                del old_value.data[relation['reverse_key']]
        
        elif relation['relation_type'].is_one_to_many():
            for item_in_bulk in old_value.bulk:
                del item_in_bulk.data[relation['reverse_key']]
        
        elif relation['relation_type'].is_many_to_one():
            old_value[relation['reverse_key']].bulk.remove(item)
            
        elif relation['relation_type'].is_many_to_many():
            for item_in_bulk in old_value:
                item_in_bulk[relation['reverse_key']].bulk.remove(item)
    
    
    def __iter_pointing_items(self, item):
        collection_set = self.__get_item_collection_set(item)
        for that_item_cls in collection_set:
            for that_item in collection_set[that_item_cls]:
                
                for key in set(that_item.data):
                    if that_item.data[key] is item:
                        yield key, that_item

    
    def item_deleted(self, item):
        """ Called when an item is deleted, for example:
        
            .. code-block:: Python
                
                item.delete()
        
        :param item: Deleted item.
        """
        # clearing the item
        # (defaults in bulk items and directs in single items)
        item.data.clear()
        
        for key, that_item in self.__iter_pointing_items(item):
            relation_type = that_item.relations[key]['relation_type']
            reverse_key = that_item.relations[key]['reverse_key']
            if reverse_key:
                if item.is_bulk_item() and relation_type.is_one_to_many():
                    for item_in_bulk in item.bulk:
                        del item_in_bulk.data[reverse_key]
                elif item.is_bulk_item() and item.bulk and \
                        relation_type.is_many_to_many():
                    
                    for this_item_in_bulk in item.bulk:
                        that_bulk = this_item_in_bulk.data[reverse_key]
    
                        for that_item_in_bulk in that_bulk.bulk:
                            
                            if that_item_in_bulk.data[key] is item:
                                that_bulk.bulk.remove(that_item_in_bulk)
            
            # defaults in bulk items and directs in single items
            del that_item.data[key]
                
        # removing from any bulks
        if item.is_single_item():
            collection_set = self.__get_item_collection_set(item)
            for that_item_cls in collection_set:
                for that_item in collection_set[that_item_cls]:
                    if that_item.is_bulk_item():
                        if item in that_item.bulk:
                            that_item.bulk.remove(item)
        else:
            item.bulk.clear()
    
    
    def item_added_to_bulk(self, item, bulk_item):
        """ Called when an item is added to a bulk item, for example:
        
            .. code-block:: Python
                
                bulk_item.add(single_item)
        
        :param item: A single item added to the `bulk_item`.
        :param bulk_item: Bulk item containing `item`.
        """
        collection_set = self.__get_item_collection_set(item)
        
        for that_item_cls in set(collection_set.keys()):
            for that_item in set(collection_set[that_item_cls]):
                if not that_item.is_single_item():
                    continue
                
                for key in set(that_item.data):
                    if that_item.data[key] is not bulk_item:
                        continue
                    
                    relation = that_item.relations[key]
                    reverse_key = that_item.relations[key]['reverse_key']
                    if not reverse_key:
                        continue
                    
                    if relation['relation_type'].is_one_to_many():
                        item.data[reverse_key] = that_item
                        
                    elif relation['relation_type'].is_many_to_many():
                        if that_item not in item[reverse_key].bulk:
                            item[reverse_key].bulk.append(that_item)
    
    
    def item_removed_from_bulk(self, item, bulk_item):
        """ Called when an item is removed from a bulk item, for example:
        
            .. code-block:: Python
                
                bulk_item.remove(single_item)
        
        :param item: A single item removed from the `bulk_item`.
        :param bulk_item: Bulk item that was containing `item`.
        """
        collection_set = self.__get_item_collection_set(item)
        
        for that_item_cls in collection_set:
            for that_item in collection_set[that_item_cls]:
                if not that_item.is_single_item():
                    continue
                
                for key in set(that_item.data):
                    if that_item.data[key] is not bulk_item:
                        continue
                    
                    relation = that_item.relations[key]
                    reverse_key = that_item.relations[key]['reverse_key']
                    
                    if reverse_key:
                        if relation['relation_type'].is_one_to_many():
                            del item.data[reverse_key]
                             
                        elif relation['relation_type'].is_many_to_many():
                            if that_item in item.data[reverse_key].bulk:
                                item.data[reverse_key].bulk.remove(that_item)

mapper = Mapper()


from itertools import chain
from .db_persister import db_persist
from .merge_models import merge_models


class AdapterBase(object):
    """ Base adapter class for all database adapters.
    Here you can see descriptions of all required methods for DB adapters
    that are going to extend this class.
    
    :ivar adapter_settings: Adapter configuration object.
    """

    #: Must be set to `True` if ORM supports composite primary keys.
    COMPOSITE_KEYS_SUPPORTED = True
    
    #: Must be set to `True` if ORM automatically updates model that is being
    #: assigned to or unassigned from another model.
    REVERSE_MODEL_AUTOUPDATE_SUPPORTED = True
    
    #: This value must be set to `True` if a model has to be first saved
    #: before its changes can be committed to database. 
    SAVE_MODEL_BEFORE_COMMIT = False
    
    #: Maximum number of models that can be pulled with one query.
    #:
    #:     .. seealso::
    #:        `batch_size` class variable of
    #:        :py:class:`~save_to_db.core.item.Item` class.
    BATCH_SIZE = 300
    
    
    def __init__(self, adapter_settings):
        self.adapter_settings = adapter_settings
        
    
    #--- general methods -------------------------------------------------------
    
    @classmethod
    def is_usable(cls, model_cls):
        """ Returns `True` if this adapter can deal with `model_cls` class.
        Here you must somehow recognize that the model class was created
        using the particular ORM library that this adapter is for.
        
        :param model_cls: ORM model class (of any library).
        :returns: Boolean value indicating whether this adapter can deal with
            the given model class.
        """
        raise NotImplementedError()
    
    
    def commit(self):
        """ Commits current transaction to database. """
        raise NotImplementedError()
    
    
    def rollback(self):
        """ Rolls back current transaction. """
        raise NotImplementedError()
    
    
    @classmethod
    def iter_fields(cls, model_cls):
        """ Returns an iteratorable of field names and their types. Foreign keys
        and relations must be ignored.
        
        :param model_cls: ORM model class for which column data are going to be
            returned.
        :returns: An iteratorable of tuples of type:
        
                *(field_name, field_type)*
            
            Where:
            
                - *field_name* is a name of an ORM model field.
                - *field_type* must be one of the fields of
                  :py:class:`~.column_type.ColumnType` enumeration class.
        """
        raise NotImplementedError()
    
    
    @classmethod
    def iter_relations(cls, model_cls):
        """ Returns an iteratorable of names of fields that reference
        other ORM models.
        
        :param model_cls: ORM model class for which relations are going to be
            iterated.
        :returns: An iteratorable of tuples of type:
        
                *(relation_field_name, relation_model_cls, relation_type,
                reverse_key)*
                
            Where:
            
            - *relation_field_name* is a field of the `model_cls` that
              references another model.
            
            - *relation_model_cls* is an ORM model class being referenced.
              
            - *relation_type* must be one of the fields of
              :py:class:`~.relation_type.RelationType` enumeration class.
            
            - *reverse_key* is the relationship name used in related model
              to reference the original model.
        """
        raise NotImplementedError()
    
    
    @classmethod
    def iter_required_fields(cls, model_cls):
        """ Returns an iteratorable of names of fields that cannot be null for
        an ORM model class:
        
            - Simple field value cannot be null if the column has a not null
              constraint.
            - Relation field cannot be null if any column used as
              foreign key (can be multiple in case of composite key) is not
              null.
        
        :param model_cls: ORM model class for which relations are going to be
            iterated.
        :returns: An iteratorable of names of fields that cannot be null.
        """
        raise NotImplementedError()
    
    
    @classmethod
    def iter_unique_field_combinations(cls, model_cls):
        """ Returns an iteratorable of unique constraints (set of
        field names) on the model.
        
        .. note::
            Relation considered to be unique if the set of columns used as
            foreign keys (can be multiple in case of composite key) has a
            unique constraint.
        
            *Relations can be unique together with other fields.*
        
        :param model_cls: ORM model class for which relations are going to be
            iterated.
        :returns: Returns an iteratorable of unique constraints.
        """
        raise NotImplementedError()
    
    
    @classmethod
    def get_table_fullname(cls, model_cls):
        """ Returns full table name with schema (if applicable) used by an
        ORM model class.
        
        :param model_cls: An ORM model class.
        :returns: Full table name with schema (if applicable). Examples:
        
                - 'public.some_table'
                - 'some_table' (no schema, e.g. SQLite database).
        """
        raise NotImplementedError()
    
    
    def get_model_cls_by_table_fullname(self, name):
        """ Return ORM model class based on full table that that model class
        uses.
        
        :param name: table full name.
        :return: ORM model class.
        """
        raise NotImplementedError()
    
    
    def iter_all_models(self):
        """ Returns an iteratorable of all known ORM model classes.
        
        :returns: An iteratorable of all known ORM model classes
        """
        raise NotImplementedError()
    
    
    #--- methods for working with items ----------------------------------------
    
    def persist(self, item):
        """ Saves item data into a database by creating or update appropriate
        database records.
        
        :param item: an instance of
            :py:class:`~save_to_db.core.item_base.ItemBase` to persist.
        :returns: Item list and corresponding ORM models as a list of lists
            (in case one item updates multiple models).
        """
        return db_persist(self, item)


    def get(self, items_and_fkeys):
        """ Accepts items and their foreign keys and gets corresponding models
        from database.
        
        .. warning::
            Items must have the same ORM model class.
        
        :param items_and_fkeys: a list of lists of type:
        
            .. code-block:: Python
            
                [
                    [item, {fkey: model, ...}],
                    ...
                ]
                
        :returns: List of created and updated models.
        """
        raise NotImplementedError()
    

    def delete(self, model):
        """ Deletes model from the database.
        
        :param model: ORM model to be deleted.
        """
        raise NotImplementedError()
    

    def create_blank_model(self, model_cls):
        """ Create empty model instance.
        
        :param model_cls: ORM model class to create.
        :returns: Newly created ORM model instance.
        """
        raise NotImplementedError()
    

    def add_related_models(self, model, fkey, related_models):
        """ Adds related models for x-to-many relationships to a parent model.
        
        :param model: Parent ORM model instance.
        :param key: Parent ORM model instance foreign key field.
        :param related_models: Referenced ORM models.
        """
        raise NotImplementedError()
    
    
    def clear_related_models(self, model, fkey):
        """ Removes related models for x-to-many relationships from a parent
        model.
        
        :param model: Parent ORM model instance.
        :param fkey: Parent ORM model instance foreign key field.
        """
        raise NotImplementedError()
    
    
    def related_x_to_many_exists(self, model, fkey):
        """ Returns `True` if `model` contains any other models with
        `field_name` relation key.
        
        :param model: Parent ORM model instance.
        :param fkey: Parent ORM model instance foreign key field.
        :returns: `True` if `model` contains any other models with
            the relation key.
        """
        raise NotImplementedError()
    

    def related_x_to_many_contains(self, model, fkey, child_models):
        """ Returns list of contained models from `child_models` by
        `model` parent model trough `field_name` foreign key field.
        
        .. note::
            The returned list must contain instances from `child_models` list,
            not other instances for the same database rows.
        
        :param model: ORM model instance.
        :param fkey: field name used for reference.
        :param child_models: list of child models.
        :returns: Returns list of contained models from `child_models` by
            `model` parent model trough `fkey` foreign key field.
        """
        raise NotImplementedError()
    
    
    @classmethod
    def get_primary_key_names(cls, model_cls):
        """ Returns tuple of primary key names.
        
        :param model_cls: ORM model class.
        :returns: Tuple of primary key names of `model_cls`.
        """
        raise NotImplementedError()
    
    
    def save_model(self, model):
        """ Saves ORM model to database.
        
        .. note::
            If possible, session must not be committed, just flushed.
        
        :param model: Parent ORM model instance.
        :param key: Parent ORM model instance foreign key field.
        """
        raise NotImplementedError()
    
    

    def execute_delete(self, model_cls, selectors, keepers):
        """ Deletes all rows in database that can be selected using `selectors`
        except those that can be selected using `keepers`.
        
        For example, suppose our `selectors` and `keepers` look like this:
        
        .. code-block:: Python
        
            selectors = [{
                'field_1': 10,
                'field_2': 20,
            }, {
                'field_1': 30,
                'field_2': 40,
            }]
            
            keepers = [{
                'field_10': 100,
                'field_20': 200,
            }, {
                'field_10': 300,
                'field_20': 400,
            }]
        
        SQL query can look like this:
        
        .. code-block:: postgres
        
            DELETE FROM model_cls_table
            WHERE
            /* selectors */
            ((field_1 = 10 AND field_2 = 20) OR
             (field_1 = 30 AND field_2 = 40))
            AND
            /* keepers */
            ((field_10 != 100 OR field_20 != 200) AND
             (field_10 != 300 OR field_20 != 400))
        
        :param model_cls: ORM model class used to work with database table.
        :param selectors: List of dictionaries with model field names as keys
            and model field values as dictionary values.
        :param keepers: Same as `selectors`.
        """
        raise NotImplementedError()
    
    
    def execute_unref(self, parent, fkey, selectors, keepers):
        """ Removes models from x-to-many field.
        
        :param parent: Parent ORM model.
        :param fkey: X-to-many foreign key field of `parent`.
        :param selectors: List of dictionaries with model field names as keys
            and model field values as dictionary values.
            
            .. seealso::
        
                `selectors` and `keepers` parameters of
                :py:meth:`~.execute_delete` method.
                
        :param keepers: Same as `selectors`.
        """
        raise NotImplementedError()
    
    
    def get_related_x_to_many(self, model, fkey):
        """ Returns iterable of many-to-many related  models.
        
        :param model: ORM model instance.
        :param fkey: field name used for reference.
        """
        return getattr(model, fkey)
    
    
    def merge_models(self, models, merge_policy=None, ignore_fields=None):
        """ Merges models into one. First model from `models` becomes the
        result model, all other models get deleted.
        
        :param models: List of ORM models (of the same class) to be merged into
            one.
        :param merge_policy: An instance of
            :py:class:`~save_to_db.core.merge_policy.MergePolicy` class.
        :param ignore_fields: List of field to ignore (not copy to result
            model) when merging.
        :returns: First model from `models` into which all other models are
            merged.
        """
        return merge_models(self, models, merge_policy=merge_policy,
                            ignore_fields=ignore_fields)
    
    
    #--- helper functions ------------------------------------------------------
    
    def pprint(self, *models):
        """ Pretty prints `models`.
        
        :param \*models: List of models to print.
        """
        if not models:
            return
        
        def repr_model(model):
            pkeys_names = self.get_primary_key_names(model.__class__)
            pkeys_values = list(str(getattr(model, pkey))
                                for pkey in pkeys_names)
            return '({},)'.format(','.join(pkeys_values))
            
        
        model_cls = None
        for model in models:
            if not (model.__class__ is model_cls):
                model_cls = model.__class__
                pkeys_names = self.get_primary_key_names(model_cls)
        
                field_names = [
                    fname for fname, _, in self.iter_fields(model_cls)
                ]
                for pk_fname in pkeys_names:
                    field_names.remove(pk_fname)
                field_names.sort()
                
                relations = {
                    fname: direction
                    for fname, _, direction, _ in self.iter_relations(model_cls)
                }
                relation_names = list(relations.keys())
                relation_names.sort()
                
                padding = len(max(chain(pkeys_names,
                                        field_names,
                                        relation_names), key=len))
                
            to_print = ['{}:'.format(model_cls.__name__)]
            for fname in chain(pkeys_names, field_names, relation_names):
                if fname in field_names or fname in pkeys_names:
                    value = getattr(model, fname)
                    no_value = value is None
                else:
                    if not relations[fname].is_x_to_many():
                        other_model = None
                        if hasattr(model, fname):
                            other_model = getattr(model, fname)
                            
                        if other_model is None:
                            value = None
                            no_value = True
                        else:
                            value = repr_model(other_model)
                            no_value = False
                    else:
                        other_models = self.get_related_x_to_many(model, fname)
                        no_value = not other_models
                        value = '[{}]'.format(
                            ','.join([repr_model(m) for m in other_models]))
                    
                to_print.append('    {:<3} {:>{padding}}: {}'.format(
                    'PK' if fname in pkeys_names else
                        '' if no_value else '::',
                    fname if no_value else fname.upper(),
                    value,
                    padding=padding))
            
            print('\n'.join(to_print))
    
    
    #--- methods for tests -----------------------------------------------------
    
    def get_all_models(self, model_cls, sort_key=None):
        """ Returns all models from database `model_cls` class.
        
        .. note::
            Used by tests only.
        
        :param model_cls: An ORM model class.
        :param sort_key: If not `None` then result is sorted using this
            function as `key` argument for `sort` method of the result list.
        :returns: List of all model instances for `model_cls` class.
        """
        raise NotImplementedError()
        


from save_to_db.exceptions import CannotMergeModels, MergeModelsNotAllowed
from save_to_db.core.item import Item
from save_to_db.core.merge_policy import MergePolicy
from save_to_db.utils.test_base import TestBase


class TestMergeModels(TestBase):
    
    def setUp(self):
        super().setUp()

        class ItemMergeOne(Item):
            model_cls = self.ModelMergeOne
        self.ItemMergeOne = ItemMergeOne
        
        class ItemMergeTwo(Item):
            model_cls = self.ModelMergeTwo
        self.ItemMergeTwo = ItemMergeTwo
        
        class ItemMergeThree(Item):
            model_cls = self.ModelMergeThree
        self.ItemMergeThree = ItemMergeThree
        
        
        class ItemMergeOneX(Item):
            model_cls = self.ModelMergeOneX
        self.ItemMergeOneX = ItemMergeOneX
        
        class ItemMergeTwoX(Item):
            model_cls = self.ModelMergeTwoX
        self.ItemMergeTwoX = ItemMergeTwoX
        
        class ItemMergeThreeX(Item):
            model_cls = self.ModelMergeThreeX
        self.ItemMergeThreeX = ItemMergeThreeX
        
        
        def x_to_one_merge(model, merged_model,
                           child_model, merged_child_model,
                           forward_foreign_key, reverse_foreign_key):
            self.merge_call_log.append([
                model, merged_model,
                child_model, merged_child_model,
                forward_foreign_key, reverse_foreign_key,
            ])
            return self.persister.merge_models(
                [child_model, merged_child_model],
                merge_policy=self.merge_policy,
                ignore_fields=[reverse_foreign_key])
            
        def one_to_many_merge(model, merged_model,
                              child_model, merged_child_model,
                              forward_foreign_key, reverse_foreign_key,
                              constaints):
            self.merge_call_log.append([
                model, merged_model,
                child_model, merged_child_model,
                
                forward_foreign_key, reverse_foreign_key,
                constaints,
            ])
            return self.persister.merge_models(
                [child_model, merged_child_model],
                merge_policy=self.merge_policy,
                ignore_fields=[reverse_foreign_key])
        
        self.merge_call_log = []
        self.merge_policy = MergePolicy(self.db_adapter, {
            self.ModelMergeOne: {
                'two_1_1': x_to_one_merge,
                'two_x_1': x_to_one_merge,
            },
            self.ModelMergeTwo: {
                'one_1_1': x_to_one_merge,
                'one_1_x': one_to_many_merge,
                'three_1_1': x_to_one_merge,
                'three_x_1': x_to_one_merge,
            },
            self.ModelMergeThree: {
                'two_1_1': x_to_one_merge,
                'two_1_x': one_to_many_merge,
            },
                             
            self.ModelMergeOneX: {
                'two_x_1': x_to_one_merge,
            },
            self.ModelMergeTwoX: {
                'one_1_x': one_to_many_merge,
                'three_x_1': x_to_one_merge,
            },
            self.ModelMergeThreeX: {
                'two_1_x': one_to_many_merge,
            },
        })
        
    
    def test_merge_propagation_1_to_1(self):
        sort_func = lambda model: model.f_integer
        
        item_one = self.ItemMergeOne(f_integer=101, f_text='text-101')
        item_two = item_one['two_1_1'](f_integer=102, f_text='text-102')
        item_two['three_1_1'](f_integer=103, f_text='text-103')
        self.persister.persist(item_one)
        
        item_one = self.ItemMergeOne(f_integer=201, f_string='string-201')
        item_two = item_one['two_1_1'](f_integer=202, f_string='string-202')
        item_two['three_1_1'](f_integer=203, f_string='string-203')
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 2)
        
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 101)
        self.assertEqual(model_one.two_1_1.f_integer, 102)
        self.assertEqual(model_one.two_1_1.three_1_1.f_integer, 103)
        
        model_one = model_ones[1]
        self.assertEqual(model_one.f_integer, 201)
        self.assertEqual(model_one.two_1_1.f_integer, 202)
        self.assertEqual(model_one.two_1_1.three_1_1.f_integer, 203)
        
        # merging
        expect_log = [
            [
                model_ones[0],
                model_ones[1],
                model_ones[0].two_1_1,
                model_ones[1].two_1_1,
                'two_1_1', 'one_1_1',
            ],
            [
                model_ones[0].two_1_1,
                model_ones[1].two_1_1,
                model_ones[0].two_1_1.three_1_1,
                model_ones[1].two_1_1.three_1_1,
                'three_1_1', 'two_1_1',
            ],
        ]
        merged_one = self.persister.merge_models(model_ones,
                                                 merge_policy=self.merge_policy)
        self.persister.db_adapter.save_model(merged_one)
        self.assertEqual(self.merge_call_log, expect_log)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(merged_one, model_ones[0])
        
        # other models should be merged
        model_twos = self.get_all_models(self.ModelMergeTwo, sort_key=sort_func)
        self.assertEqual(len(model_twos), 1)
        model_threes = self.get_all_models(self.ModelMergeThree,
                                           sort_key=sort_func)
        self.assertEqual(len(model_threes), 1)
        
        # checking the model
        self.assertEqual(merged_one.f_integer, 101)
        self.assertEqual(merged_one.f_text, 'text-101')
        self.assertEqual(merged_one.f_string, 'string-201')
        self.assertEqual(merged_one.two_1_1.f_integer, 102)
        self.assertEqual(merged_one.two_1_1.f_text, 'text-102')
        self.assertEqual(merged_one.two_1_1.f_string, 'string-202')
        self.assertEqual(merged_one.two_1_1.three_1_1.f_integer, 103)
        self.assertEqual(merged_one.two_1_1.three_1_1.f_text, 'text-103')
        self.assertEqual(merged_one.two_1_1.three_1_1.f_string, 'string-203')
    
    
    def test_merge_no_propagation_1_to_1(self):
        sort_func = lambda model: model.f_integer
        
        item_one = self.ItemMergeOne(f_integer=101, f_text='text-101')
        item_two = item_one['two_1_1'](f_integer=102, f_text='text-102')
        item_two['three_1_1'](f_integer=103, f_text='text-103')
        self.persister.persist(item_one)
        
        item_one = self.ItemMergeOne(f_integer=201, f_string='string-201')
        item_two = item_one['two_1_1'](f_integer=202, f_string='string-202')
        item_two['three_1_1'](f_integer=203, f_string='string-203')
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 2)
        
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 101)
        self.assertEqual(model_one.two_1_1.f_integer, 102)
        self.assertEqual(model_one.two_1_1.three_1_1.f_integer, 103)
        
        model_one = model_ones[1]
        self.assertEqual(model_one.f_integer, 201)
        self.assertEqual(model_one.two_1_1.f_integer, 202)
        self.assertEqual(model_one.two_1_1.three_1_1.f_integer, 203)
        
        # merging
        with self.assertRaises(CannotMergeModels):
            self.persister.merge_models(model_ones)
        
    
    def test_merge_propagation_x_to_1(self):
        sort_func = lambda model: model.f_integer
        
        item_one = self.ItemMergeOne(f_integer=101, f_text='text-101')
        item_two = item_one['two_x_1'](f_integer=102, f_text='text-102')
        item_two['three_x_1'](f_integer=103, f_text='text-103')
        self.persister.persist(item_one)
        
        item_one = self.ItemMergeOne(f_integer=201, f_string='string-201')
        item_two = item_one['two_x_1'](f_integer=202, f_string='string-202')
        item_two['three_x_1'](f_integer=203, f_string='string-203')
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 2)
        
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 101)
        self.assertEqual(model_one.two_x_1.f_integer, 102)
        self.assertEqual(model_one.two_x_1.three_x_1.f_integer, 103)
        
        model_one = model_ones[1]
        self.assertEqual(model_one.f_integer, 201)
        self.assertEqual(model_one.two_x_1.f_integer, 202)
        self.assertEqual(model_one.two_x_1.three_x_1.f_integer, 203)
        
        # merging
        expect_log = [
            [
                model_ones[0],
                model_ones[1],
                model_ones[0].two_x_1,
                model_ones[1].two_x_1,
                'two_x_1', 'one_1_x',
            ],
            [
                model_ones[0].two_x_1,
                model_ones[1].two_x_1,
                model_ones[0].two_x_1.three_x_1,
                model_ones[1].two_x_1.three_x_1,
                'three_x_1', 'two_1_x',
            ],
        ]
        merged_one = self.persister.merge_models(model_ones,
                                                 merge_policy=self.merge_policy)
        self.persister.db_adapter.save_model(merged_one)
        self.assertEqual(self.merge_call_log, expect_log)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(merged_one, model_ones[0])
        
        # other models should be merged
        model_twos = self.get_all_models(self.ModelMergeTwo, sort_key=sort_func)
        self.assertEqual(len(model_twos), 1)
        model_threes = self.get_all_models(self.ModelMergeThree,
                                           sort_key=sort_func)
        self.assertEqual(len(model_threes), 1)
        
        # checking the model
        self.assertEqual(merged_one.f_integer, 101)
        self.assertEqual(merged_one.f_text, 'text-101')
        self.assertEqual(merged_one.f_string, 'string-201')
        self.assertEqual(merged_one.two_x_1.f_integer, 102)
        self.assertEqual(merged_one.two_x_1.f_text, 'text-102')
        self.assertEqual(merged_one.two_x_1.f_string, 'string-202')
        self.assertEqual(merged_one.two_x_1.three_x_1.f_integer, 103)
        self.assertEqual(merged_one.two_x_1.three_x_1.f_text, 'text-103')
        self.assertEqual(merged_one.two_x_1.three_x_1.f_string, 'string-203')
    
    
    def test_merge_no_propagation_x_to_1(self):
        sort_func = lambda model: model.f_integer
        
        item_one = self.ItemMergeOne(f_integer=101, f_text='text-101')
        item_two = item_one['two_x_1'](f_integer=102, f_text='text-102')
        item_two['three_x_1'](f_integer=103, f_text='text-103')
        self.persister.persist(item_one)
        
        item_one = self.ItemMergeOne(f_integer=201, f_string='string-201')
        item_two = item_one['two_x_1'](f_integer=202, f_string='string-202')
        item_two['three_x_1'](f_integer=203, f_string='string-203')
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 2)
        
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 101)
        self.assertEqual(model_one.two_x_1.f_integer, 102)
        self.assertEqual(model_one.two_x_1.three_x_1.f_integer, 103)
        
        model_one = model_ones[1]
        self.assertEqual(model_one.f_integer, 201)
        self.assertEqual(model_one.two_x_1.f_integer, 202)
        self.assertEqual(model_one.two_x_1.three_x_1.f_integer, 203)
        
        # merging
        with self.assertRaises(CannotMergeModels):
            self.persister.merge_models(model_ones)
    
    
    def __structure_models_for_propagation_1_to_x(self, model_threes):
        sort_func = lambda model: model.f_integer
        
        structure = []
        for model_three in model_threes:
            entry_three = {
                'f_string': model_three.f_string,
                'f_integer': model_three.f_integer,
                'entries': []
            }
            structure.append(entry_three)
            for model_two in self.get_related_x_to_many(
                        model_three, 'two_1_x', sort_key=sort_func):
                entry_two = {
                    'f_string': model_two.f_string,
                    'f_integer': model_two.f_integer,
                    'entries': []
                }
                entry_three['entries'].append(entry_two)
                for model_one in self.get_related_x_to_many(
                        model_two, 'one_1_x',  sort_key=sort_func):
                    entry_one = {
                        'f_string': model_one.f_string,
                        'f_integer': model_one.f_integer,
                        'entries': []
                    }
                    entry_two['entries'].append(entry_one)
        
        return structure
                    
    
    def __generate_models_for_propagation_1_to_x(self):
        # Two structurees will be merged (three->two->one):
        #     /-1
        #    1--0
        # 0--|
        #    0--0
        #     \-0
        #
        # Where models must that must be merged mark as 1.
        
        sort_func = lambda model: model.f_string
        
        bulk_three = self.ItemMergeThreeX.Bulk()
        for x in range(1, 3):
            item_three = bulk_three.gen(
                f_integer=x*100,
                f_string='x: {}'.format(x))
            
            for y in range(1, 3):
                item_two = item_three['two_1_x'].gen(
                    f_integer=x*100 + y*10,
                    f_string='x: {}; y: {}'.format(x, y))
                
                for z in range(1, 3):
                    item_two['one_1_x'].gen(
                        f_integer=x*100 + y*10 + z,
                        f_string='x: {}; y: {}; z: {}'.format(x, y, z))

        self.persister.persist(bulk_three)
        model_threes = self.get_all_models(self.ModelMergeThreeX,
                                           sort_key=sort_func)
        self.assertEqual(len(model_threes), 2)
        
        model_three_1 = model_threes[0]
        model_three_2 = model_threes[1]
        
        # making some models to merge
        two_1_x_one = self.get_related_x_to_many(model_three_1, 'two_1_x',
                                                 sort_key=sort_func)
        two_1_x_two = self.get_related_x_to_many(model_three_2, 'two_1_x',
                                                 sort_key=sort_func)
        self.assertEqual(len(two_1_x_one), 2)
        self.assertEqual(len(two_1_x_two), 2)
        
        two_1_x_one[0].f_integer = two_1_x_two[0].f_integer
        self.persister.db_adapter.save_model(two_1_x_one[0])
        
        one_1_x_one = self.get_related_x_to_many(two_1_x_one[0], 'one_1_x',
                                                 sort_key=sort_func)
        one_1_x_two = self.get_related_x_to_many(two_1_x_two[0], 'one_1_x',
                                                 sort_key=sort_func)
        self.assertEqual(len(one_1_x_one), 2)
        self.assertEqual(len(one_1_x_two), 2)
        
        one_1_x_one[0].f_integer = one_1_x_two[0].f_integer
        self.persister.db_adapter.save_model(one_1_x_one[0])
        
    
    def test_merge_propagation_1_to_x(self):
        sort_func = lambda model: model.f_string
        self.__generate_models_for_propagation_1_to_x()
        
        # merging models
        model_threes = self.get_all_models(self.ModelMergeThreeX,
                                           sort_key=sort_func)
        self.assertEqual(len(model_threes), 2)
          
          
        two_1_x_one = self.get_related_x_to_many(model_threes[0], 'two_1_x',
                                                 sort_key=sort_func)[0]
        two_1_x_two = self.get_related_x_to_many(model_threes[1], 'two_1_x',
                                                 sort_key=sort_func)[0]
        one_1_x_one = self.get_related_x_to_many(two_1_x_one, 'one_1_x',
                                                 sort_key=sort_func)[0]
        one_1_x_two = self.get_related_x_to_many(two_1_x_two, 'one_1_x',
                                                 sort_key=sort_func)[0]
                                                   
        # merging
        # (for django models can be different after merge because of deletion)
        expect_log = [
            [
                model_threes[0].f_string,
                model_threes[1].f_string,
                two_1_x_one.f_string,
                two_1_x_two.f_string,
                'two_1_x', 'three_x_1',
                [{'three_x_1', 'f_integer'}]
            ],
            [
                two_1_x_one.f_string,
                two_1_x_two.f_string,
                one_1_x_one.f_string,
                one_1_x_two.f_string,
                'one_1_x', 'two_x_1',
                [{'two_x_1', 'f_integer'}]
            ],
        ]
        merged_three = self.persister.merge_models(
            model_threes, merge_policy=self.merge_policy)
        self.persister.db_adapter.save_model(merged_three)
        
        for entry in self.merge_call_log:
            for i in range(4):
                entry[i] = entry[i].f_string
        self.assertEqual(self.merge_call_log, expect_log)
        
        model_threes = self.get_all_models(self.ModelMergeThreeX,
                                           sort_key=sort_func)
        self.assertEqual(len(model_threes), 1)
        self.assertEqual(merged_three, model_threes[0])
         
        structure = self.__structure_models_for_propagation_1_to_x(
            model_threes)
        
        expected = [{
            'entries': [{
                'entries': [{
                    'entries': [],
                    'f_integer': 121,
                    'f_string': 'x: 1; y: 2; z: 1'
                }, {
                    'entries': [],
                    'f_integer': 122,
                    'f_string': 'x: 1; y: 2; z: 2'
                }],
                'f_integer': 120,
                'f_string': 'x: 1; y: 2'
            }, {
                'entries': [{
                    'entries': [],
                    'f_integer': 112,
                    'f_string': 'x: 1; y: 1; z: 2'
                }, {
                    'entries': [],
                    'f_integer': 211,
                    'f_string': 'x: 1; y: 1; z: 1'
                }, {
                    'entries': [],
                    'f_integer': 212,
                    'f_string': 'x: 2; y: 1; z: 2'
                }],
                'f_integer': 210,
                'f_string': 'x: 1; y: 1'
            }, {
                'entries': [{
                    'entries': [],
                    'f_integer': 221,
                    'f_string': 'x: 2; y: 2; z: 1'
                }, {
                    'entries': [],
                    'f_integer': 222,
                    'f_string': 'x: 2; y: 2; z: 2'
                }],
                'f_integer': 220,
                'f_string': 'x: 2; y: 2'
            }],
            'f_integer': 100,
            'f_string': 'x: 1'
        }]
        
        self.assertEqual(structure, expected)
        
        
    def test_merge_no_propagation_1_to_x(self):
        self.__generate_models_for_propagation_1_to_x()
        
        # merging models
        model_threes = self.get_all_models(self.ModelMergeThreeX)
        self.assertEqual(len(model_threes), 2)
        
        # for now we do not check before exception
        with self.assertRaises(Exception):
            self.persister.merge_models(model_threes)
    
    
    def test_ignore_regular_field(self):
        sort_func = lambda model: model.f_integer
        
        bulk_one = self.ItemMergeOne.Bulk()
        bulk_one.gen(f_integer=1, f_float=2.0)
        bulk_one.gen(f_integer=2, f_float=3.0, f_text='text', f_string='string')
        
        self.persister.persist(bulk_one)
        models_one = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(models_one), 2)
        
        merged_model = self.persister.merge_models(models_one,
                                                   ignore_fields=['f_string'])
        self.persister.db_adapter.save_model(merged_model)
        
        models_one = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(models_one), 1)
        
        model_one = models_one[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_float, 2)
        self.assertEqual(model_one.f_text, 'text')
        self.assertIsNone(model_one.f_string)
        
    
    def test_merging_not_allowed(self):
        sort_func = lambda model: model.f_integer
        
        #--- x-to-one allowed --------------------------------------------------
        # (CannotMergeModels exception will be raised)
        bulk_one = self.ItemMergeOne.Bulk()
        bulk_one.gen(f_integer=1,
                     two_1_1=self.ItemMergeTwo(f_integer=2))
        bulk_one.gen(f_integer=3,
                     two_1_1=self.ItemMergeTwo(f_integer=4))
        self.persister.persist(bulk_one)
        
        models_one = self.get_all_models(self.ModelMergeOne)
        self.assertEqual(len(models_one), 2)
        
        models_two = self.get_all_models(self.ModelMergeTwo)
        self.assertEqual(len(models_two), 2)
        
        with self.assertRaises(CannotMergeModels):
            self.persister.merge_models(models_one)
        
        #--- x-to-one not allowed ----------------------------------------------
        merge_policy = MergePolicy(self.db_adapter, {
            self.ModelMergeOne: {
                'two_1_1': False
            }
        })
        with self.assertRaises(MergeModelsNotAllowed):
            self.persister.merge_models(models_one, merge_policy=merge_policy)
        
        #--- x-to-many allowed -------------------------------------------------
        bulk_two = self.ItemMergeTwo.Bulk()
        bulk_two.gen(f_integer=101,
                     one_1_x=[self.ItemMergeOne(f_integer=102)])
        bulk_two.gen(f_integer=103,
                     one_1_x=[self.ItemMergeOne(f_integer=104)])
        self.persister.persist(bulk_two)
        
        models_one = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(models_one), 4)
        
        models_two = self.get_all_models(self.ModelMergeTwo, sort_key=sort_func)
        self.assertEqual(len(models_two), 4)
        
        models_two = [m for m in models_two if m.f_integer in [101, 103]]
        
        merged_model = self.persister.merge_models(models_two)
        
        models_one = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(models_one), 4)
        
        models_two = self.get_all_models(self.ModelMergeTwo, sort_key=sort_func)
        self.assertEqual(len(models_two), 3)
        
        self.assertEqual(merged_model.f_integer, 101)
        one_1_x = self.get_related_x_to_many(merged_model, 'one_1_x',
                                             sort_key=sort_func)
        self.assertEqual(len(one_1_x), 2)
        
        self.assertEqual(one_1_x[0].f_integer, 102)
        self.assertEqual(one_1_x[1].f_integer, 104)
        
        #--- x-to many not allowed ---------------------------------------------
        bulk_two = self.ItemMergeTwoX.Bulk()
        bulk_two.gen(f_integer=201,
                     one_1_x=[self.ItemMergeOneX(f_integer=202)])
        bulk_two.gen(f_integer=203,
                     one_1_x=[self.ItemMergeOneX(f_integer=202)])
        self.persister.persist(bulk_two)
        
        models_one = self.get_all_models(self.ModelMergeOneX, sort_key=sort_func)
        self.assertEqual(len(models_one), 2)
        
        models_two = self.get_all_models(self.ModelMergeTwoX, sort_key=sort_func)
        self.assertEqual(len(models_two), 2)
        
        merge_policy = MergePolicy(self.db_adapter, {
            self.ModelMergeTwoX: {
                'one_1_x': False
            }
        })
        models_two = [m for m in models_two if m.f_integer in [201, 203]]
        with self.assertRaises(MergeModelsNotAllowed):
            self.persister.merge_models(models_two, merge_policy=merge_policy)
        
    
    def test_merging_1_to_1_first_none(self):
        sort_func = lambda model: model.f_integer
        
        bulk_one = self.ItemMergeOne.Bulk()
        bulk_one.gen(f_integer=1)  # two_1_1 is None
        bulk_one.gen(f_integer=2,
                     two_1_1=self.ItemMergeTwo(f_integer=3))
        self.persister.persist(bulk_one)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 2)
        model_twos = self.get_all_models(self.ModelMergeTwo, sort_key=sort_func)
        self.assertEqual(len(model_twos), 1)
        
        merged_model = self.persister.merge_models(model_ones)
        
        model_ones = self.get_all_models(self.ModelMergeOne, sort_key=sort_func)
        self.assertEqual(len(model_ones), 1)
        model_twos = self.get_all_models(self.ModelMergeTwo, sort_key=sort_func)
        self.assertEqual(len(model_twos), 1)
        
        self.assertEqual(merged_model.f_integer, 1)
        self.assertEqual(merged_model.two_1_1.f_integer, 3)
        
        


import os
import django
from django.core.management import call_command

from save_to_db.adapters.django_adapter import DjangoAdapter
from save_to_db.core.persister import Persister

from . import project
from .project import settings
from .django_test_base import DjangoTestBase

from .. import register_test
from ..test_fields_and_relations import TestFieldsAndRelations
from ..test_constraints import TestConstraints
from ..test_item_cls_manager import TestItemClsManager
from ..test_merge_policy import TestMergeModels, TestPolicySetup
from ..test_item_persist import (TestAdapter,
                                 TestExceptions,
                                 TestItemHooks,
                                 TestResolveModel,
                                 TestModelDeleter,
                                 TestNorewrite,
                                 TestPersist,
                                 TestReplace,
                                 TestSimpleConfig)
from ..test_item_use import (TestBasicUse,
                             TestItemDump,
                             TestItemProcess,
                             TestItemSetup)
from ..test_scope import TestScope
from ..test_select import TestSelect
from ..test_signals import TestSignals
from ..test_mapper import TestMapper


# setting up Django and loading models
settings.INSTALLED_APPS = tuple(
    '{}.{}'.format(project.__name__, app) for app in settings.INSTALLED_APPS)
os.environ.setdefault(
    'DJANGO_SETTINGS_MODULE',
    '{}.settings'.format(project.__name__))
django.setup()
call_command('migrate', verbosity=0)
from .project.orm_only import models



AdapterPrefix = 'Django'
persister = Persister(DjangoAdapter({}))

models = {
    'ModelFieldTypes': models.ModelFieldTypes,
    
    'ModelGeneralOne': models.ModelGeneralOne,
    'ModelGeneralTwo': models.ModelGeneralTwo,

    'ModelConstraintsOne': models.ModelConstraintsOne,
    'ModelConstraintsTwo': models.ModelConstraintsTwo,
    'ModelConstraintsThree': models.ModelConstraintsThree,
    'ModelConstraintsFour': models.ModelConstraintsFour,
    'ModelConstraintsFive': models.ModelConstraintsFive,
    'ModelConstraintsSix': models.ModelConstraintsSix,
    'ModelConstraintsSelf': models.ModelConstraintsSelf,
    
    'ModelInvalidFieldNames': models.ModelInvalidFieldNames,

    'ModelManyRefsOne': models.ModelManyRefsOne,
    'ModelManyRefsTwo': models.ModelManyRefsTwo,
    
    'ModelAutoReverseOne': models.ModelAutoReverseOne,
    'ModelAutoReverseTwoA': models.ModelAutoReverseTwoA,
    'ModelAutoReverseTwoB': models.ModelAutoReverseTwoB,
    'ModelAutoReverseThreeA': models.ModelAutoReverseThreeA,
    'ModelAutoReverseThreeB': models.ModelAutoReverseThreeB,
    'ModelAutoReverseFourA': models.ModelAutoReverseFourA,
    'ModelAutoReverseFourB': models.ModelAutoReverseFourB,
    
    'ModelMergeOne': models.ModelMergeOne,
    'ModelMergeTwo': models.ModelMergeTwo,
    'ModelMergeThree': models.ModelMergeThree,
    'ModelMergeOneX': models.ModelMergeOneX,
    'ModelMergeTwoX': models.ModelMergeTwoX,
    'ModelMergeThreeX': models.ModelMergeThreeX,
}

for test_cls in (TestFieldsAndRelations,
                 TestConstraints,
                 TestItemClsManager,
                 TestMergeModels,
                 TestPolicySetup,
                 TestAdapter,
                 TestExceptions,
                 TestItemHooks,
                 TestResolveModel,
                 TestModelDeleter,
                 TestNorewrite,
                 TestPersist,
                 TestReplace,
                 TestSimpleConfig,
                 TestBasicUse,
                 TestItemDump,
                 TestItemProcess,
                 TestItemSetup,
                 TestItemDump,
                 TestScope,
                 TestSelect,
                 TestSignals,
                 TestMapper):
    
    register_test(AdapterPrefix, DjangoTestBase,
                  test_cls, models, persister)


import datetime
from collections import OrderedDict

from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase


class TestItemProcess(TestBase):
    
    ModelFieldTypes = None
    ModelGeneralOne = None
    ModelGeneralTwo = None
    ModelAutoReverseOne = None
    ModelAutoReverseTwoA = None
    ModelAutoReverseTwoB = None
    ModelAutoReverseThreeA = None
    ModelAutoReverseThreeB = None
    ModelAutoReverseFourA = None
    ModelAutoReverseFourB = None
    
    
    @classmethod
    def setup_models(cls, aliased=False):
        cls.item_cls_manager.clear()
        
        class ItemFieldTypes(Item):
            model_cls = cls.ModelFieldTypes
            conversions = {
                'decimal_separator': ',',
            }
        cls.ItemFieldTypes = ItemFieldTypes
        
        dct_1 = {'model_cls': cls.ModelGeneralOne}
        dct_2 = {'model_cls': cls.ModelGeneralTwo}
        
        if aliased:
            dct_1.update({
                'aliases': {
                    'alias_1': 'two_1_1__f_integer',
                    'alias_2': 'two_x_x__one_1_1__f_integer',
                }
            })
            dct_2.update({
                'aliases': {
                    'alias_1__post': 'f_string',
                    'alias_2__post': 'one_x_1__alias_1',
                    'alias_3__post': 'one_x_x__two_1_x__alias_1__post',
                }
            })
        
        cls.ItemGeneralOne = type('ItemGeneralOne', (Item,), dct_1)
        cls.ItemGeneralTwo = type('ItemGeneralTwo', (Item,), dct_2)
        
        cls.ItemGeneralOne.complete_setup()
        cls.ItemGeneralTwo.complete_setup()
    
        
    def test_field_convesions(self):
        self.setup_models()
        
        #--- simple conversion ---
        item = self.ItemFieldTypes(
            binary_1 = 'binary data',
            string_1 = 1000,
            text_1 = 2000,
            integer_1 = '10',
            boolean_1 = 'TRUE',
            float_1 = '1.120,30',  # with comma as decimal separator and a dot
            date_1 = '2000-10-30',
            time_1 = '20:30:40',
            datetime_1 = '2000-10-30 20:30:40')
        
        item.process()
        
        expected_value = {
            'binary_1': b'binary data',
            'boolean_1': True,
            'date_1': datetime.date(2000, 10, 30),
            'datetime_1': datetime.datetime(2000, 10, 30, 20, 30, 40),
            'float_1': 1120.3,
            'integer_1': 10,
            'string_1': '1000',
            'text_1': '2000',
            'time_1': datetime.time(20, 30, 40)
        }
        for key, value in expected_value.items():
            self.assertEqual(item[key], value)
            
        item.process()  # second processing does nothing
        
        #--- conversions with relations ---
        item = self.ItemGeneralOne(f_integer='10',
                                   two_x_1__f_integer='20',
                                   two_x_x__f_integer='30')
        item['two_x_x'].gen(f_integer='40')
        item.process()
        
        self.assertEqual(item['f_integer'], 10)
        self.assertEqual(item['two_x_1__f_integer'], 20)
        self.assertEqual(item['two_x_x__f_integer'], 30)
        self.assertEqual(item['two_x_x'][0]['f_integer'], 40)
        
    
    def test_inject_nullables(self):
        self.setup_models()
        
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            # using normal fields
            nullables = ['f_integer', 'f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            # using relations
            nullables = ['one_1_1', 'one_x_1', 'one_1_x', 'one_x_x']
        
        # no overwrite for normal fields
        item = ItemGeneralOne(f_integer='10', f_string='20', f_text='30')
        item.process()
        self.assertEqual(item['f_integer'], 10)
        self.assertEqual(item['f_string'], '20')
        self.assertEqual(item['f_text'], '30')
        
        # normal fields nullables added
        item = ItemGeneralOne()
        item.process()
        self.assertIsNone(item['f_integer'])
        self.assertIsNone(item['f_string'])
        with self.assertRaises(KeyError):
            item['f_text']
        
        # no overwrite for relations
        item = ItemGeneralTwo(one_1_1__f_integer='10',
                              one_1_1__f_string='20',
                              one_x_1__f_integer='10',
                              one_x_1__f_string='20')
        item['one_1_x'].gen(f_integer='10', f_string='20')
        item['one_x_x'].gen(f_integer='10', f_string='20')
        item.process()
        
        self.assertEqual(item['one_1_1']['f_integer'], 10)
        self.assertEqual(item['one_1_1']['f_string'], '20')
        self.assertEqual(item['one_x_1']['f_integer'], 10)
        self.assertEqual(item['one_x_1']['f_string'], '20')
        
        for key in ['one_1_x', 'one_x_x']:
            for item_one in item[key]:
                self.assertEqual(item_one['f_integer'], 10)
                self.assertEqual(item_one['f_string'], '20')
        
        # relation fields nullables added
        item = ItemGeneralTwo()
        item.process()
        
        for key in ['one_1_1', 'one_x_1', 'one_1_x', 'one_x_x']:
            self.assertIn(key, item)
        
        self.assertIsNone(item['one_1_1'])
        self.assertIsNone(item['one_x_1'])
        
        self.assertEqual(len(item['one_1_x']), 0)
        self.assertEqual(len(item['one_x_x']), 0)
        
    
    def test_aliases_process(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            aliases = {
                'alias_1': 'two_x_1__f_integer',
                'alias_2': 'two_x_x__f_integer',
            }
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            aliases = {
                'alias_1__post': 'f_float',
                'alias_2__post': 'one_x_1__alias_1',
                'alias_3__post': 'one_x_x__two_x_1__alias_1__post',
            }

        ItemGeneralOne.complete_setup()
        ItemGeneralTwo.complete_setup()


        # processing individual fields
        value = ItemGeneralOne.process_field('alias_1', '1')
        self.assertEqual(value, 1)
        
        value = ItemGeneralTwo.process_field('alias_1__post', '3.3')
        self.assertEqual(value, 3.3)
        value = ItemGeneralTwo.process_field('alias_2__post', '4')
        self.assertEqual(value, 4)
        value = ItemGeneralTwo.process_field('alias_3__post', '5.5')
        self.assertEqual(value, 5.5)
        
        # processing items
        item = ItemGeneralOne(alias_1='1',
                              alias_2='2',
                              two_1_1__alias_1__post='3.3',
                              two_1_1__alias_2__post='5',
                              two_1_1__alias_3__post='5.5')
        item.process()
        
        self.assertEqual(item['alias_1'], 1)
        self.assertEqual(item['alias_2'], 2)
        self.assertEqual(item['two_x_1__f_integer'], 1)
        self.assertEqual(item['two_x_x__f_integer'], 2)
        
        two_1_1 = item['two_1_1']
        self.assertEqual(two_1_1['alias_1__post'], 3.3)
        self.assertEqual(two_1_1['alias_2__post'], 5)
        self.assertEqual(two_1_1['alias_3__post'], 5.5)
        
        self.assertEqual(two_1_1['f_float'], 3.3)
        self.assertEqual(two_1_1['one_x_1__two_x_1__f_integer'], 5)
        self.assertEqual(two_1_1['one_x_x__two_x_1__f_float'], 5.5)
        
        bulk = ItemGeneralOne.Bulk()
        # generating same item as before
        bulk.gen(alias_1='1',
                 alias_2='2',
                 two_1_1__alias_1__post='3.3',
                 two_1_1__alias_2__post='5',
                 two_1_1__alias_3__post='5.5')
        
        # two_1_1__alias_2__post
        bulk['two_1_1__one_x_1__two_x_1__f_integer'] = '6'
        # two_x_1__one_x_x__two_x_1__f_float
        bulk['two_x_1__alias_3__post'] = '6.6'
        bulk.process()
        
        self.assertEqual(bulk['two_1_1__alias_2__post'], 6)
        self.assertEqual(bulk['two_x_1__alias_3__post'], 6.6)
        
        self.assertEqual(bulk['two_1_1__one_x_1__two_x_1__f_integer'], 6)
        self.assertEqual(bulk['two_x_1__one_x_x__two_x_1__f_float'], 6.6)
        
        item = bulk[0]
        
        self.assertEqual(item['alias_1'], 1)
        self.assertEqual(item['alias_2'], 2)
        self.assertEqual(item['two_x_1__f_integer'], 1)
        self.assertEqual(item['two_x_x__f_integer'], 2)
        
        self.assertEqual(item['two_1_1__alias_1__post'], 3.3)
        self.assertEqual(item['two_1_1__alias_2__post'], 5)
        self.assertEqual(item['two_1_1__alias_3__post'], 5.5)
        self.assertEqual(item['two_x_1__alias_3__post'], 6.6)
        
        self.assertEqual(item['two_1_1__f_float'], 3.3)
        self.assertEqual(item['two_1_1__one_x_1__two_x_1__f_integer'], 5)
        self.assertEqual(item['two_1_1__one_x_x__two_x_1__f_float'], 5.5)
        self.assertEqual(item['two_x_1__one_x_x__two_x_1__f_float'], 6.6)
    
    
    def test_inject_bulk_defaults(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            # using normal fields
            nullables = ['f_integer', 'f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            # using relations
            nullables = ['one_x_1', 'one_x_x']
        
        # default regular fields
        # (first default fields must be injected only then nullables)
        bulk = ItemGeneralOne.Bulk(f_integer='1000', f_text='text-value')
        bulk.gen(f_integer='10', f_float='20.30')
        bulk.gen(f_string='str-value', f_float='20.30')
        bulk.process()
        expect = {
            'bulk': [
                {
                    'item': {
                        'f_float': 20.3,
                        'f_integer': 10,
                        'f_string': None,
                        'f_text': 'text-value'
                    }
                },
                {
                    'item': {
                        'f_float': 20.3,
                        'f_integer': 1000,
                        'f_string': 'str-value',
                        'f_text': 'text-value'
                    }
                }
            ],
            'defaults': {
                'f_integer': 1000,
                'f_text': 'text-value'
            }
        }
        self.assertEqual(bulk.to_dict(), expect)
        
        # default relations
        # many-to-many using item
        default_two_x_1 = ItemGeneralTwo(f_integer='10', f_float='20.30')
        default_two_x_x = ItemGeneralTwo.Bulk()
        default_two_x_x.gen(f_integer='20')
        default_two_x_x.gen(f_float='40.50')
        
        bulk = ItemGeneralOne.Bulk(two_x_1=default_two_x_1,
                                   two_x_x=default_two_x_x)
        bulk.gen(two_x_1__f_integer=40)
        bulk.gen(two_x_x__f_text='text-value')
        bulk.process()
        expect = {
            'bulk': [
                # first item: `bulk.gen(two_x_1__f_integer=40)`
                # + nullables + defaults
                {
                    'id': 1,
                    'item': {
                        'f_integer': None,
                        'f_string': None,
                        'two_x_1': {
                            'item': {
                                'f_integer': 40,
                                'one_1_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'one_x_1': None,
                                'one_x_x': {
                                    'bulk': [],
                                    'defaults': {}
                                }
                            }
                        },
                        # `default_two_x_x = ItemGeneralTwo.Bulk()` with two
                        # items:
                        # `default_two_x_x.gen(f_integer='20')` and
                        # `default_two_x_x.gen(f_float='40.50')` + nullables
                        'two_x_x': {
                            'bulk': [
                                {
                                    'item': {
                                        'f_integer': 20,
                                        'one_x_1': None,
                                        'one_x_x': {
                                            'bulk': [
                                                {
                                                    'id': 1
                                                }
                                            ],
                                            'defaults': {}
                                        }
                                    }
                                },
                                {
                                    'item': {
                                        'f_float': 40.5,
                                        'one_x_1': None,
                                        'one_x_x': {
                                            'bulk': [
                                                {
                                                    'id': 1
                                                }
                                            ],
                                            'defaults': {}
                                        }
                                    }
                                }
                            ],
                            'defaults': {}
                        }
                    }
                },
                {
                    'id': 2,
                    'item': {
                        'f_integer': None,
                        'f_string': None,
                        # `ItemGeneralTwo(f_integer='10', f_float='20.30')`
                        # + nullables
                        'two_x_1': {
                            'item': {
                                'f_float': 20.3,
                                'f_integer': 10,
                                'one_1_x': {
                                    'bulk': [
                                        {
                                            'id': 2
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'one_x_1': None,
                                'one_x_x': {
                                    'bulk': [],
                                    'defaults': {}
                                }
                            }
                        },
                        # already existed
                        'two_x_x': {
                            'bulk': [],
                            'defaults': {
                                'f_text': 'text-value'
                            }
                        }
                    }
                }
            ],
            'defaults': {
                'two_x_1': {
                    'item': {
                        'f_float': 20.3,
                        'f_integer': 10,
                        'one_x_1': None,
                        'one_x_x': {
                            'bulk': [],
                            'defaults': {}
                        }
                    }
                },
                'two_x_x': {
                    'bulk': [
                        {
                            'item': {
                                'f_integer': 20,
                                'one_x_1': None,
                                'one_x_x': {
                                    'bulk': [],
                                    'defaults': {}
                                }
                            }
                        },
                        {
                            'item': {
                                'f_float': 40.5,
                                'one_x_1': None,
                                'one_x_x': {
                                    'bulk': [],
                                    'defaults': {}
                                }
                            }
                        }
                    ],
                    'defaults': {}
                }
            }
        }
        self.assertTrue(bulk.to_dict(), expect)
        
        # many-to-many using list value
        bulk = ItemGeneralOne.Bulk(two_x_1=default_two_x_1,
                                   two_x_x=default_two_x_x.as_list())
        bulk.gen(two_x_1__f_integer=40)
        bulk.gen(two_x_x__f_text='text-value')
        bulk.process()
        self.assertTrue(bulk.to_dict(), expect)
    
    
    def test_plain_defaults_injection(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            defaults ={
                'f_integer': '100',
                'f_boolean': 'true',
                'f_string': lambda item: 'INT: {}'.format(item['f_integer']),
            }
            
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        
        item = ItemGeneralOne(f_integer='400')
        item.process()
        self.assertEqual(item['f_integer'], 400)
        self.assertIs(item['f_boolean'], True)
        self.assertEqual(item['f_string'], 'INT: 400')
        
        item = ItemGeneralOne(f_boolean='false')
        item.process()
        self.assertEqual(item['f_integer'], 100)
        self.assertIs(item['f_boolean'], False)
        self.assertEqual(item['f_string'], 'INT: 100')
        
        
    def test_item_defaults_injection(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        # regular default values -----------------------------------------------
        default_bulk = ItemGeneralOne.Bulk()
        default_bulk.add(ItemGeneralOne(f_float='19.1'),
                         ItemGeneralOne(f_float='19.2'))
        ItemGeneralTwo.defaults.update({
            'one_1_x': default_bulk,  # as bulk
            'one_x_x': [
                ItemGeneralOne(f_float='99.1'),
                ItemGeneralOne(f_float='99.2'),
            ],  # as list
        })
        
        item_two = ItemGeneralTwo(f_float='2')
        item_two.process()
        
        # making sure that all items were processed (converted to float)
        self.assertEqual(item_two['f_float'], 2)
        self.assertEqual(item_two['one_1_x'][0]['f_float'], 19.1)
        self.assertEqual(item_two['one_1_x'][1]['f_float'], 19.2)
        self.assertEqual(item_two['one_x_x'][0]['f_float'], 99.1)
        self.assertEqual(item_two['one_x_x'][1]['f_float'], 99.2)
        
        # must not be overwritten with default
        item_two = ItemGeneralTwo(f_float='2.2',
                                  one_1_x=[ItemGeneralOne(f_float='19.12'),
                                           ItemGeneralOne(f_float='19.22')])
        item_two['one_x_x'].add(ItemGeneralOne(f_float='99.12'),
                                ItemGeneralOne(f_float='99.22'))
        item_two.process()
        self.assertEqual(item_two['f_float'], 2.2)
        self.assertEqual(item_two['one_1_x'][0]['f_float'], 19.12)
        self.assertEqual(item_two['one_1_x'][1]['f_float'], 19.22)
        self.assertEqual(item_two['one_x_x'][0]['f_float'], 99.12)
        self.assertEqual(item_two['one_x_x'][1]['f_float'], 99.22)
        
        # function default values ----------------------------------------------
        def gen_bulk_item(_):
            bulk_two = ItemGeneralTwo.Bulk()
            bulk_two.gen(f_float='99.1')
            bulk_two.gen(f_float='99.2')
            return bulk_two
        
        default_bulk = ItemGeneralTwo.Bulk()
        default_bulk.add(ItemGeneralTwo(f_float='19.1'),
                         ItemGeneralTwo(f_float='19.2'))
        ItemGeneralTwo.defaults.clear()
        ItemGeneralOne.defaults.update({
            'two_1_x': lambda _: default_bulk.as_list(),  # as list
            'two_x_x': gen_bulk_item,  # as item
        })
        
        default_bulk.add(ItemGeneralTwo(f_float='19.1'))
        
        item_one = ItemGeneralOne(f_float='1')
        
        item_one.process()
        
        # making sure that all items were processed (converted to float)
        self.assertEqual(item_one['f_float'], 1)
        self.assertEqual(item_one['two_1_x'][0]['f_float'], 19.1)
        self.assertEqual(item_one['two_1_x'][1]['f_float'], 19.2) 
        self.assertEqual(item_one['two_x_x'][0]['f_float'], 99.1)
        self.assertEqual(item_one['two_x_x'][1]['f_float'], 99.2)
        
        # must not be overwritten with default
        item_one = ItemGeneralOne(f_float='1.2',
                                  two_1_x=[ItemGeneralTwo(f_float='19.12'),
                                           ItemGeneralTwo(f_float='19.22')])
        item_one['two_x_x'].add(ItemGeneralTwo(f_float='99.12'),
                                ItemGeneralTwo(f_float='99.22'))
        item_one.process()
        self.assertEqual(item_one['f_float'], 1.2)
        self.assertEqual(item_one['two_1_x'][0]['f_float'], 19.12)
        self.assertEqual(item_one['two_1_x'][1]['f_float'], 19.22)
        self.assertEqual(item_one['two_x_x'][0]['f_float'], 99.12)
        self.assertEqual(item_one['two_x_x'][1]['f_float'], 99.22)
    
    
    def test_defaults_injection_order_in_bulk(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        bulk_one = ItemGeneralOne.Bulk()
        # making sure that default iteration will be in a certain order
        bulk_one.data = OrderedDict()
        bulk_one['two_x_1__one_x_1__f_string'] = 'default-1'
        bulk_one['two_x_1__f_string'] = 'default-2'
        bulk_one.gen(f_string='test-item')
        bulk_one.process()
        
        def assertItem(bulk_one):
            self.assertEqual(bulk_one['two_x_1__one_x_1__f_string'],
                             'default-1')
            self.assertEqual(bulk_one['two_x_1__f_string'], 'default-2')
            
            item_one = bulk_one[0]
            # value from the item itself
            self.assertEqual(item_one['f_string'], 'test-item')
            # "two_x_1" must contain this value from default ...
            self.assertEqual(item_one['two_x_1']['f_string'], 'default-2')
            # ... and contain an item from default
            one_x_1 = item_one['two_x_1']['one_x_1']
            self.assertEqual(one_x_1['f_string'], 'default-1')
        
        assertItem(bulk_one)
        
        bulk_one = ItemGeneralOne.Bulk()
        bulk_one.data = OrderedDict()
        # different order
        bulk_one['two_x_1__f_string'] = 'default-2'
        bulk_one['two_x_1__one_x_1__f_string'] = 'default-1'
        bulk_one.gen(f_string='test-item')
        bulk_one.process()
        
        # but result is the same
        assertItem(bulk_one)
    
    
    def test_remove_null_fields(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            remove_null_fields = ['f_integer', 'f_string',
                                  'two_1_x', 'two_1_1']
            
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            remove_null_fields = ['f_integer', 'f_string',
                                  'one_1_x', 'one_1_1']
            nullables = ['f_string', 'one_1_1']
        
        item_one = ItemGeneralOne(f_integer=None,
                                  f_string='string-value',
                                  two_1_x=[],
                                  two_1_1=None,
                                  two_x_x=[],
                                  two_x_1=None)
        item_one.process()
        # remove_null_fields must not be present
        self.assertNotIn('f_integer', item_one)
        self.assertNotIn('two_1_x', item_one)
        self.assertNotIn('two_1_1', item_one)
        # f_string field has not None value
        self.assertIn('f_string', item_one)
        
        item_two = ItemGeneralTwo(f_integer=None,
                                  f_string=None,
                                  one_1_x=[ItemGeneralOne()],
                                  one_1_1=None)
        item_two.process()
        # remove_null_fields must not be present
        self.assertNotIn('f_integer', item_two)
        self.assertNotIn('two_1_x', item_two)
        # f_string and one_1_1 are nullables
        self.assertIn('f_string', item_two)
        self.assertIn('one_1_1', item_two)
        # one_1_x has value
        self.assertIn('one_1_x', item_two)
        
        